unit Unit6;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  fverinfo, ExtCtrls, StdCtrls, ShellAPI, Registry, Buttons;


type
  TAboutForm = class(TForm)
    Bevel1: TBevel;
    btnOk: TButton;
    Label1: TLabel;
    ProductVersion: TLabel;
    Label4: TLabel;
    Label2: TLabel;
    Bevel3: TBevel;
    WinVersion: TLabel;
    Homepage: TLabel;
    lblHomepage: TLabel;
    Image1: TImage;
    PhysMem: TLabel;
    SpeedButton1: TSpeedButton;
    procedure HomepageClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure GetOSInfo;
    procedure InitializeCaptions;
    procedure SpeedButton1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  AboutForm: TAboutForm;

implementation

Uses Unit1;
{$R *.DFM}

procedure OpenURL(Url: string);
var
  ts: string;
begin
  with TRegistry.Create do
    try
      rootkey := HKEY_CLASSES_ROOT;
      OpenKey('\htmlfile\shell\open\command', False);
      try
        ts := ReadString('');
      except
        ts := '';
      end;
      CloseKey;
    finally
      Free;
    end;
  if ts = '' then Exit;
  // remove quotes and commandline parameters
  ts := Copy(ts, Pos('"', ts) + 1, Length(ts));
  ts := Copy(ts, 1, Pos('"', ts) - 1);
  ShellExecute(0, 'open', PChar(ts), PChar(url), nil, SW_SHOW);
end;


procedure TAboutForm.HomepageClick(Sender: TObject);
begin
  OpenURL(Homepage.Caption);
end;

procedure TAboutForm.FormCreate(Sender: TObject);
begin
  Self.Caption:=MudMapperForm.GetCaption('AboutForm','Caption',Self.Caption);
  ProductVersion.Caption:=MudMapperForm.GetCaption('AboutForm','ProductVersion',ProductVersion.Caption);
  WinVersion.Caption:=MudMapperForm.GetCaption('AboutForm','WinVersion',WinVersion.Caption);
  lblHomepage.Caption:=MudMapperForm.GetCaption('AboutForm','lblHomepage',lblHomepage.Caption);
  btnOk.Caption:=MudMapperForm.GetCaption('AboutForm','btnOk',btnOk.Caption);
  InitializeCaptions;
end;

procedure TAboutForm.GetOSInfo;
var
  Platform: string;
  BuildNumber: Integer;
begin
  case Win32Platform of
    VER_PLATFORM_WIN32_WINDOWS:
      begin
        Platform := 'Windows 95';
        BuildNumber := Win32BuildNumber and $0000FFFF;
      end;
    VER_PLATFORM_WIN32_NT:
      begin
        Platform := 'Windows NT';
        BuildNumber := Win32BuildNumber;
      end;
      else
      begin
        Platform := 'Windows';
        BuildNumber := 0;
      end;
  end;
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) or
    (Win32Platform = VER_PLATFORM_WIN32_NT) then
  begin
    if Win32CSDVersion = '' then
      WinVersion.Caption := Format('%s %d.%d (Build %d)', [Platform, Win32MajorVersion,
        Win32MinorVersion, BuildNumber])
    else
      WinVersion.Caption := Format('%s %d.%d (Build %d: %s)', [Platform, Win32MajorVersion,
        Win32MinorVersion, BuildNumber, Win32CSDVersion]);
  end
  else
    WinVersion.Caption := Format('%s %d.%d', [Platform, Win32MajorVersion,
      Win32MinorVersion])
end;

procedure TAboutForm.InitializeCaptions;
var
  MS: TMemoryStatus;
begin
  GetOSInfo;
  MS.dwLength := SizeOf(TMemoryStatus);
  GlobalMemoryStatus(MS);
  PhysMem.Caption := 'Memory available: '+FormatFloat('#,###" KB"', MS.dwTotalPhys div 1024);
end;

procedure TAboutForm.SpeedButton1Click(Sender: TObject);
var s:String;
begin
s:=MudMapperForm.CheckUpdate;
if s='err'
then
Application.MessageBox(PChar('     .'),
           PChar(fProductName),MB_ICONINFORMATION+MB_OK+MB_SYSTEMMODAL)
else
if s<>fProductVersion
then
Application.MessageBox(PChar('   - '+s+'.'),
           PChar(fProductName),MB_ICONINFORMATION+MB_OK+MB_SYSTEMMODAL)
else
Application.MessageBox(PChar('    .'),
           PChar(fProductName),MB_ICONINFORMATION+MB_OK+MB_SYSTEMMODAL)
end;

end.
